"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import math
import numpy as np
import warnings

import autograd.numpy as anp
from autograd import grad
import scipy.optimize as sopt
import functools
from sklearn.utils import check_array, resample
from sklearn.linear_model import LinearRegression

from .base import _BaseLiNGAM
from .bootstrap import BootstrapResult


class ABICLiNGAM(_BaseLiNGAM):
    """Implementation of ABIC-LiNGAM Algorithm. [1]_
    Original code: https://github.com/Yoshimitsu-try/ABIC_LiNGAM

    References
    ----------
    .. [1] Y. Morinishi and S. Shimizu. Differentiable causal discovery of
       linear non-Gaussian acyclic models under unmeasured confounding.
       Transactions on Machine Learning Research (TMLR), 2025.
    """

    def __init__(
        self,
        beta=1.0,
        lam=0.05,
        acyc_order=None,
        seed=0,
        max_outer=100,
        tol_h=1e-8,
        min_causal_effect=0.05,
        min_error_covariance=0.05,
        rho_max=1e16,
        inner_start=1,
        inner_growth=1,
        inner_tol=1e-4,
    ):
        """Construct a ABICLiNGAM model.

        Parameters
        ----------
        beta : float, optional (default=1.0)
            Power in residual loss, i.e., ||r||^(2*beta)
        lam : float, optional (default=0.05)
            The weight of the regularization term.
        acyc_order : int or None, optional (default=None)
            Order of the truncated series for acyclicity penalty. If None, defaults to the number of variables.
        min_causal_effect : float, optional (default=0.05)
            Threshold for detecting causal edge.
            Causal edges with absolute values of causal effects less than ``min_causal_effect`` are excluded.
        min_error_covariance : float, optional (default=0.05)
            Threshold for detecting error covariances.
            Error covariances with absolute values less than ``min_error_covariance`` are excluded.
        seed : int, optional (default=0)
            Seed for the random number generator.
        max_outer : int, optional (default=100)
            Maximum number of outer iterations.
        tol_h : float, optional (default=1e-8)
            Tolerance for acyclicity penalty to stop.
        rho_max : float, optional (default=1e16)
            Maximum value for Augmented Lagrangian penalty parameter rho.
        inner_start : int, optional (default=1)
            Initial number of inner refinement steps.
        inner_growth : int, optional (default=1)
            Growth of inner refinement steps per outer iteration.
        inner_tol : float, optional (default=1e-4)
            Tolerance for inner loop convergence.
        """
        # Check parameters
        if beta <= 0.0:
            raise ValueError("beta must be positive.")
        if lam < 0.0:
            raise ValueError("lam must be non-negative.")
        if acyc_order is not None:
            if not isinstance(acyc_order, int):
                raise TypeError("acyc_order must be an integer or None.")
            if acyc_order < 1:
                raise ValueError("acyc_order must be >= 1.")
        if min_causal_effect < 0.0:
            raise ValueError("min_causal_effect must be non-negative.")
        if min_error_covariance < 0.0:
            raise ValueError("min_error_covariance must be non-negative.")
        if max_outer < 1:
            raise ValueError("max_outer must be at least 1.")
        if tol_h <= 0.0:
            raise ValueError("tol_h must be positive.")
        if rho_max <= 0.0:
            raise ValueError("rho_max must be positive.")
        if inner_start < 1:
            raise ValueError("inner_start must be at least 1.")
        if inner_growth < 0:
            raise ValueError("inner_growth must be non-negative.")
        if inner_tol <= 0.0:
            raise ValueError("inner_tol must be positive.")

        self._beta = float(beta)
        self._lam = float(lam)
        self._acyc_order = acyc_order
        self._min_causal_effect = float(min_causal_effect)
        self._min_error_covariance = float(min_error_covariance)
        self._seed = int(seed)
        self._max_outer = int(max_outer)
        self._tol_h = float(tol_h)
        self._rho_max = float(rho_max)
        self._inner_start = int(inner_start)
        self._inner_growth = int(inner_growth)
        self._inner_tol = float(inner_tol)

        super().__init__()

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Observed data matrix.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self._X = anp.asarray(check_array(X))
        d = self._X.shape[1]

        # Random number generator
        self._rng = np.random.default_rng(self._seed)

        # Check parameters
        if d < 2:
            raise ValueError("Data must have at least two variables (features).")

        # Initialize parameters
        B = anp.array(self._rng.uniform(-0.5, 0.5, size=(d, d)))
        L = anp.array(self._rng.uniform(-0.05, 0.05, size=(d, d)))
        lower_mask = anp.array(np.tril(np.ones((d, d)), k=-1))
        L = L * lower_mask
        L = L + L.T
        L = L - anp.diag(anp.diag(L))
        D = anp.diag(anp.diag(anp.cov(self._X.T)))

        rho, alpha, h_prev = 1.0, 0.0, np.inf
        inner_cap = self._inner_start
        penalty_fn = self._bow_penalty

        bounds = self._build_bounds()  # no prior knowledge
        objective = functools.partial(self._objective)
        gradient = grad(objective)

        for _ in range(self._max_outer):
            B_new, L_new, D_new = None, None, None
            h_new = None

            while rho < self._rho_max:
                B_new = B.copy()
                L_new = L.copy()
                D_new = D.copy()

                # inner refinement
                for _ in range(inner_cap):
                    B_old, L_old, D_old = B_new, L_new, D_new
                    Z = self._pseudo(B_new, L_new + D_new)

                    theta0 = anp.concatenate([anp.ravel(B_new), anp.ravel(L_new)])
                    res = sopt.minimize(
                        self._objective,
                        theta0,
                        args=(rho, alpha, Z, penalty_fn),
                        method="L-BFGS-B",
                        jac=gradient,
                        bounds=bounds,
                        options={"disp": False},
                    )

                    B_new = anp.reshape(res.x[: d * d], (d, d))
                    L_new = anp.reshape(res.x[d * d :], (d, d))
                    L_new = L_new + L_new.T
                    L_new = L_new - anp.diag(anp.diag(L_new))

                    # refresh diagonal noise from residuals (different expression)
                    diag_vals = [
                        anp.var(self._X[:, j] - anp.dot(self._X, B_new[:, j]))
                        for j in range(d)
                    ]
                    D_new = anp.diag(anp.array(diag_vals))

                    # convergence of inner loop
                    delta = anp.sum(anp.abs(B_old - B_new)) + anp.sum(
                        anp.abs((L_old + D_old) - (L_new + D_new))
                    )
                    if float(delta) < self._inner_tol:
                        break

                h_new = self._acyclicity_penalty(B_new) + penalty_fn(B_new, L_new)

                # penalty schedule
                if float(h_new) < 0.25 * float(h_prev):
                    break
                else:
                    rho *= 10.0

            # AL update
            B, L, D = B_new.copy(), L_new.copy(), D_new.copy()
            h_prev = h_new
            alpha = alpha + rho * h_prev
            inner_cap += self._inner_growth

            if float(h_prev) <= self._tol_h or rho >= self._rho_max:
                break

        self._B = anp.where(anp.abs(B) < self._min_causal_effect, 0.0, B)
        self._omega = anp.where(
            anp.abs(L + D) < self._min_error_covariance, 0.0, (L + D)
        )

        # Merge coefficient matrix and error covariance matrix
        omega_copied = self._omega.copy()
        anp.fill_diagonal(omega_copied, 0.0)
        omega_copied[anp.abs(omega_copied) > 0] = anp.nan
        self._adjacency_matrix = self._B.T.copy()
        self._adjacency_matrix[anp.isnan(omega_copied)] = anp.nan

        # Estimate causal order from coefficient matrix
        self._causal_order = self._causal_order_from_adjacency_matrix(self._B.T)

        return self

    def bootstrap(self, X, n_sampling=100):
        """Bootstrap sampling to assess variability of estimates.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Observed data matrix.
        n_sampling : int, optional (default=100)
            Number of bootstrap samples.

        Returns
        -------
        Bs : array-like, shape (n_sampling, n_features, n_features)
            Bootstrap samples of estimated adjacency matrices.
        Omegas : array-like, shape (n_sampling, n_features, n_features)
            Bootstrap samples of estimated error covariance matrices.
        """
        X = anp.asarray(X)
        d = X.shape[1]

        adjacency_matrices = anp.zeros((n_sampling, d, d))
        Bs = anp.zeros((n_sampling, d, d))
        Omegas = anp.zeros((n_sampling, d, d))
        total_effects = anp.zeros((n_sampling, d, d))
        index = anp.arange(X.shape[0])
        resampled_indices = []

        for i in range(n_sampling):
            resampled_X, resampled_index = resample(X, index)
            resampled_indices.append(resampled_index)
            self.fit(resampled_X)
            adjacency_matrices[i] = self._adjacency_matrix
            Bs[i] = self._B
            Omegas[i] = self._omega

            # Calculate total effects
            for c, from_ in enumerate(self._causal_order):
                for to in self._causal_order[c + 1 :]:
                    if True in np.isnan(self._adjacency_matrix[from_]):
                        total_effects[i, to, from_] = np.nan
                    else:
                        total_effects[i, to, from_] = self.estimate_total_effect(
                            resampled_X, from_, to
                        )

        return ABICBootstrapResult(
            adjacency_matrices,
            Bs,
            Omegas,
            total_effects,
            resampled_indices=resampled_indices,
        )

    def _acyclicity_penalty(self, W, K=None):
        """Smooth acyclicity surrogate written as a truncated series:
            h(W) = sum_{k=1..K} trace((W∘W)^k) / k!
        where ∘ is Hadamard product. K defaults to d.
        This differs in form/implementation from common "M=I+..." variants
        and avoids any custom VJP; autograd handles the gradient.

        Parameters
        ----------
        W : array-like, shape (d, d)
            Directed adjacency matrix.
        K : int or None
            Order of the truncated series. If None, defaults to d.

        Returns
        -------
        penalty : float
            Value of the acyclicity penalty.
        """
        d = W.shape[0]
        if K is None:
            K = self._acyc_order or d
        A = W * W
        Ak = anp.eye(d)
        acc = 0.0
        for k in range(1, K + 1):
            Ak = anp.dot(Ak, A)
            acc = acc + anp.trace(Ak) / float(math.factorial(k))
        return acc

    @staticmethod
    def _bow_penalty(W1, W2):
        """Bow-freeness surrogate in an alternative form:
            || W1 ∘ W2 ||_F^2 / |W1|

        Parameters
        ----------
        W1 : array-like, shape (d, d)
            Directed adjacency matrix.
        W2 : array-like, shape (d, d)
            Bidirected adjacency matrix.

        Returns
        -------
        penalty : float
            Value of the bow-freeness penalty.
        """
        A = W1 * W2
        return anp.sum(A * A) / A.size

    def _objective(self, theta, rho, alpha, Z, penalty_fn):
        """Augmented Lagrangian objective. All pieces are auto-diff compatible.

        Parameters
        ----------
        theta : array-like, shape (d*d + d*d,)
            Concatenated parameter vector [vec(B), vec(L)], where L is strictly lower
            triangular part to be mirrored to form a symmetric matrix.
        rho : float
            The weight of the penalty term
        alpha : float
            The Lagrange multiplier
        Z : list of array-like, shape (n, d)
            Pseudo-variables for bidirected part.
        penalty_fn : callable
            Structural penalty function on (B, L_sym).

        Returns
        -------
        obj : float
            Value of the objective function.
        """

        n, d = self._X.shape

        # unpack and enforce symmetry for the bidirected part
        B = anp.reshape(theta[: d * d], (d, d))
        L = anp.reshape(theta[d * d :], (d, d))
        L = L + L.T
        L = L - anp.diag(anp.diag(L))  # zero diagonal

        # LS(theta) term (with generalized power 2*beta)
        LS = 0.0
        for j in range(d):
            r = self._X[:, j] - anp.dot(self._X, B[:, j]) - anp.dot(Z[j], L[:, j])
            LS = LS + 0.5 / n * (anp.linalg.norm(r) ** (2 * self._beta))

        # structural constraints
        h = self._acyclicity_penalty(B) + penalty_fn(B, L)
        aug = 0.5 * rho * (h**2) + alpha * h

        # smooth L0-ish (tanh-like) regularization on theta
        s = anp.log(n) * anp.abs(theta)
        t = (anp.exp(s) - 1) / (anp.exp(s) + 1)
        return LS + aug + self._lam * anp.sum(t)

    def _build_bounds(self, levels=None, exogenous=(), w_range=4.0):
        """Create L-BFGS-B bounds for theta

        Parameters
        ----------
        levels : list[list[index]] or None
            Prior knowledge about variable ordering. Each sublist represents a level,
            where variables in earlier levels cannot have incoming edges from later levels.
        exogenous : array-like, shape (index, ...)
            Indices of exogenous variables (no incoming edges).
        w_range : float
            Range for weights (default: 4.0). Bounds are set to [-w_range, w_range].

        Returns
        -------
        bounds : list[tuple[float, float]]
            Bounds for each parameter in theta = [vec(B), vec(L)].
        """
        d = self._X.shape[1]

        if levels is None:
            levels = [[i for i in range(d)]]

        tier = {v: t for t, group in enumerate(levels) for v in group}
        exo = set(exogenous)

        # Directed bounds B: start wide, zero diag, then forbid backward edges
        B_lo = -w_range * np.ones((d, d))
        B_hi = +w_range * np.ones((d, d))
        np.fill_diagonal(B_lo, 0.0)
        np.fill_diagonal(B_hi, 0.0)
        for i in range(d):
            for j in range(d):
                if i == j:
                    continue
                if tier[i] > tier[j]:
                    B_lo[i, j] = 0.0
                    B_hi[i, j] = 0.0

        # Bidirected bounds L (we optimize lower-tri; upper/diag fixed zero)
        L_lo = -w_range * np.ones((d, d))
        L_hi = +w_range * np.ones((d, d))
        for i in range(d):
            for j in range(d):
                if i <= j or (i in exo) or (j in exo):
                    L_lo[i, j] = 0.0
                    L_hi[i, j] = 0.0

        bounds_B = np.c_[B_lo.reshape(-1), B_hi.reshape(-1)].tolist()
        bounds_L = np.c_[L_lo.reshape(-1), L_hi.reshape(-1)].tolist()
        return bounds_B + bounds_L

    def _pseudo(self, B, omega):
        """
        Build pseudo-variables Z using solves instead of explicit inversion.

        Parameters
        ----------
        B : array-like, shape (d, d)
            Current estimate of directed adjacency matrix.
        omega : array-like, shape (d, d)
            Current estimate of error covariance matrix.

        Returns
        -------
        Z : list of array-like, shape (n, d)
            Returns a list Z such that Z[j] has a zero column at j (shape (n, d)).
        """
        d = B.shape[0]
        eps = self._X - anp.dot(self._X, B)
        Z = [None] * d
        for j in range(d):
            idx = [k for k in range(d) if k != j]
            omega_ = omega[anp.ix_(idx, idx)]
            Zij = anp.linalg.solve(omega_, eps[:, idx].T).T  # (n, d-1)
            Zj = anp.insert(Zij, j, 0.0, axis=1)  # (n, d)
            Z[j] = Zj
        return Z

    def _causal_order_from_adjacency_matrix(self, matrix, threshold=1e-8):
        """Estimate causal order from adjacency matrix using Kahn's algorithm."""
        d = matrix.shape[0]
        # 有向グラフの隣接リストと入次数を作成
        adj = {i: set() for i in range(d)}
        indegree = [0] * d
        for i in range(d):
            for j in range(d):
                if i != j and abs(matrix[i, j]) > threshold:
                    adj[i].add(j)
                    indegree[j] += 1

        # 入次数0のノードから順に並べる
        order = []
        queue = [i for i in range(d) if indegree[i] == 0]
        while queue:
            n = queue.pop(0)
            order.append(n)
            for m in adj[n]:
                indegree[m] -= 1
                if indegree[m] == 0:
                    queue.append(m)

        if len(order) != d:
            raise ValueError("The adjacency matrix contains a cycle (not a DAG).")

        return order[::-1]

    def estimate_total_effect(self, X, from_index, to_index):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        # Check parameters
        X = check_array(X)

        # Check confounders
        if True in np.isnan(self._adjacency_matrix[from_index]):
            warnings.warn(
                f"The estimated causal effect may be incorrect because "
                f"the source variable (from_index={from_index}) is influenced by confounders."
            )
            return np.nan

        # from_index + parents indices
        parents = np.where(np.abs(self._adjacency_matrix[from_index]) > 0)[0]
        predictors = [from_index]
        predictors.extend(parents)

        # Estimate total effect
        lr = LinearRegression()
        lr.fit(X[:, predictors], X[:, to_index])

        return lr.coef_[0]

    @property
    def coefficient_matrix_(self):
        """Estimated coefficient matrix.

        Returns
        -------
        coefficient_matrix_ : array-like, shape (n_features, n_features)
            The coefficient matrix B of fitted model, where
            n_features is the number of features.
        """
        return self._B

    @property
    def error_covariance_matrix_(self):
        """Estimated error covariance matrix.

        Returns
        -------
        error_covariance_matrix_ : array-like, shape (n_features, n_features)
            The error covariance matrix Omega of fitted model, where
            n_features is the number of features.
        """
        return self._omega


class ABICBootstrapResult(BootstrapResult):
    """The result of bootstrapping for Time series algorithm."""

    def __init__(
        self,
        adjacency_matrices,
        coefficient_matrices,
        error_covariance_matrices,
        total_effects,
        resampled_indices=None,
    ):
        """Construct a BootstrapResult.

        Parameters
        ----------
        adjacency_matrices : array-like, shape (n_sampling)
            The adjacency matrix list by bootstrapping.
        coefficient_matrices : array-like, shape (n_sampling)
            The coefficient matrix list by bootstrapping.
        error_covariance_matrices : array-like, shape (n_sampling)
            The error covariance matrix list by bootstrapping.
        total_effects : array-like, shape (n_sampling)
            The total effects list by bootstrapping.
        resampled_indices : list of array-like, shape (n_sampling), optional (default=None)
            The list of resampled indices used in bootstrapping.
        """
        super().__init__(
            adjacency_matrices, total_effects, resampled_indices=resampled_indices
        )
        self._coefficient_matrices = coefficient_matrices
        self._error_covariance_matrices = error_covariance_matrices

    @property
    def coefficient_matrices_(self):
        """The coefficient matrix list by bootstrapping.

        Returns
        -------
        coefficient_matrices_ : array-like, shape (n_sampling)
            The coefficient matrix list, where ``n_sampling`` is
            the number of bootstrap sampling.
        """
        return self._coefficient_matrices

    @property
    def error_covariance_matrices_(self):
        """The error covariance matrix list by bootstrapping.

        Returns
        -------
        error_covariance_matrices_ : array-like, shape (n_sampling)
            The error covariance matrix list, where ``n_sampling`` is
            the number of bootstrap sampling.
        """
        return self._error_covariance_matrices



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import numpy as np
import statsmodels.api as sm
from sklearn.utils import check_array
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression, LogisticRegression, Lasso, LassoLarsIC

from .base import _BaseLiNGAM
from .utils import bic_select_logistic_l1


class mLiNGAM(_BaseLiNGAM):
    """Implementation of the Missingness-LiNGAM (m-LiNGAM) Algorithm
    """
    def __init__(
        self,
        random_state=None,
        prior_knowledge=None,
        apply_prior_knowledge_softly=False,
        missingness_knowledge=None,
    ):
        """Construct an m-LiNGAM model.

        Parameters
        ----------
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        prior_knowledge : array-like, shape (n_features, n_features), optional (default=None)
            Prior knowledge used for causal discovery, where ``n_features`` is the number of features.

            The elements of prior knowledge matrix are defined as follows [1]_:

            * ``0`` : :math:`x_i` does not have a directed path to :math:`x_j`
            * ``1`` : :math:`x_i` has a directed path to :math:`x_j`
            * ``-1`` : No prior knowledge is available to know if either of the two cases above (0 or 1) is true.
        apply_prior_knowledge_softly : boolean, optional (default=False)
            If True, apply prior knowledge softly.
        """
        super().__init__(random_state)
        self._Aknw = prior_knowledge
        self._apply_prior_knowledge_softly = apply_prior_knowledge_softly
        self._missingness_knowledge = missingness_knowledge
        self._missingness_mechanisms_parents = {}  # Dictionary. Key: partially observed variable index, Value: corresponding missingness mechanism's parents indexes (list)
        self._missingness_mechanisms_coef = {}
        self.n_features = None
        self._descendants_mechanisms = None
        self._missing_to_index = None
        self._intercept = None

        if self._Aknw is not None:
            self._Aknw = check_array(self._Aknw)
            self._Aknw = np.where(self._Aknw < 0, np.nan, self._Aknw)

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features. Missing values may be represented as NaN.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parameters
        X = check_array(X, ensure_all_finite='allow-nan')
        self.n_features = X.shape[1]

        # Check prior knowledge
        if self._Aknw is not None:
            if (self.n_features, self.n_features) != self._Aknw.shape:
                raise ValueError(
                    "The shape of prior knowledge must be (n_features, n_features)"
                )
            else:
                # Extract all partial orders in prior knowledge matrix
                if not self._apply_prior_knowledge_softly:
                    self._partial_orders = self._extract_partial_orders(self._Aknw)

        # Causal discovery
        U = np.arange(self.n_features)
        K = []
        X_ = np.copy(X)

        # Discover missingness mechanisms
        missing_mask = np.isnan(X_)
        columns_with_missing = np.any(missing_mask, axis=0)
        missing_column_indices = np.where(columns_with_missing)[0]
        self._missing_to_index = {i : m for m, i in enumerate(missing_column_indices)}
        self._descendants_mechanisms = -1 * np.ones([len(missing_column_indices), self.n_features])

        R = {i : missing_mask[:, i] for i in missing_column_indices}
        for k in R.keys():
            # Find the parent nodes of the missingness mechanism
            available_rows = ~np.any(np.isnan(np.delete(X_, k, axis=1)), axis=1)

            X_lreg = np.delete(X_, k, axis=1)[available_rows]
            scaler = StandardScaler()
            X_lreg = scaler.fit_transform(X_lreg)

            best_coef, _, _, _, _ = bic_select_logistic_l1(X_lreg, R[k][available_rows], Cs=50, max_iter=1000)

            self._missingness_mechanisms_parents[k] = list(np.arange(self.n_features)[np.where(best_coef != 0)])
            self._missingness_mechanisms_parents[k] = [idx if idx < k else idx + 1 for idx in self._missingness_mechanisms_parents[k]]
            available_rows = ~np.any(np.isnan(X_[:, self._missingness_mechanisms_parents[k]]), axis=1)

            if len(self._missingness_mechanisms_parents[k]) == 0:
                clf = LogisticRegression(
                    penalty=None,
                    solver='lbfgs',
                    max_iter=1000,
                    fit_intercept=False
                )
                # independent_vars = np.ones_like(R[k])
                independent_vars = np.ones_like(R[k]).reshape(-1, 1)
                clf.fit(independent_vars, R[k][available_rows])
                self._missingness_mechanisms_coef[k] = np.concatenate([clf.coef_.ravel()])
            else:
                clf = LogisticRegression(
                    penalty='l2',
                    C=0.5,
                    solver='lbfgs',
                    max_iter=1000,
                    fit_intercept=True
                )
                independent_vars = X_[:, self._missingness_mechanisms_parents[k]][available_rows]
                clf.fit(independent_vars, R[k][available_rows])
                self._missingness_mechanisms_coef[k] = np.concatenate([clf.intercept_, clf.coef_.ravel()])

        # Estimate causal order
        X_top = X_.copy()

        for _ in range(len(U)):
            m = self._search_causal_order_top_down(X_top, U, min_samples=self.n_features + 1)
            for i in U:
                if i != m:
                    X_top[:, i][~np.isnan(X[:, [i, m]]).any(axis=1)] = self._residual(X_top[:, i][~np.isnan(X[:, [i, m]]).any(axis=1)], X_top[:, m][~np.isnan(X[:, [i, m]]).any(axis=1)])
            K.append(m)
            U = U[U != m]
            # Update partial orders
            if (self._Aknw is not None) and (not self._apply_prior_knowledge_softly):
                self._partial_orders = self._partial_orders[
                    self._partial_orders[:, 0] != m
                ]

        self._causal_order = K

        # Estimate adjacency matrix
        if self._Aknw is not None:
            pk = self._Aknw.copy()
            np.fill_diagonal(pk, 0)

        B = np.zeros([X.shape[1], X.shape[1]], dtype="float64")
        c = np.zeros(X.shape[1], dtype="float64")
        for i in range(len(self._causal_order)):
            target = self._causal_order[i]
            predictors = self._causal_order[:i]

            # Exclude variables specified in no_path with prior knowledge
            if self._Aknw is not None:
                predictors = [p for p in predictors if pk[target, p] != 0]

            while len(predictors) > 0:
                n_pred = len(predictors)
                # Calculate coefficients of the original scale
                involved_variables = set([target] + predictors)
                num_involved = len(involved_variables)
                while True:
                    for j in involved_variables:
                        if j in self._missingness_mechanisms_parents.keys():
                            involved_variables = involved_variables.union(set(self._missingness_mechanisms_parents[j]))
                    if num_involved == len(involved_variables):
                        break
                    else:
                        num_involved = len(involved_variables)
                X_m = X[~np.any(np.isnan(X[:, sorted(involved_variables)]), axis=1)]
                sample_weight = np.prod([1 / (1 - self._logistic_prediction(X_m[:, self._missingness_mechanisms_parents[i]], self._missingness_mechanisms_coef[i])) for i in [target] + predictors if i in self._missingness_mechanisms_parents.keys()], axis=0)

                # Standardize X
                scaler = StandardScaler()
                X_std = scaler.fit_transform(X_m)

                # Pruning with Adaptive Lasso
                lr = Lasso(alpha=0.1)
                lr.fit(X_std[:, predictors], X_std[:, target], sample_weight=sample_weight)
                weight = np.power(np.abs(lr.coef_), 1.0)
                reg = LassoLarsIC(criterion="bic")
                reg.fit(X_std[:, predictors] * weight, X_std[:, target])
                pruned_idx = np.abs(reg.coef_ * weight) > 0.0

                pred = np.array(predictors)
                predictors = list(pred[pruned_idx])

                if len(predictors) == n_pred:
                    break

            # Calculate coefficients of the original scale
            involved_variables = set([target] + predictors)
            num_involved = len(involved_variables)
            while True:
                for j in involved_variables:
                    if j in self._missingness_mechanisms_parents.keys():
                        involved_variables = involved_variables.union(set(self._missingness_mechanisms_parents[j]))
                if num_involved == len(involved_variables):
                    break
                else:
                    num_involved = len(involved_variables)
            X_m = X[~np.any(np.isnan(X[:, sorted(involved_variables)]), axis=1)]
            sample_weight = np.prod([1 / (1 - self._logistic_prediction(X_m[:, self._missingness_mechanisms_parents[i]], self._missingness_mechanisms_coef[i])) for i in [target] + predictors if i in self._missingness_mechanisms_parents.keys()], axis=0)

            if len(predictors) == 0:
                X_design = np.ones((X_m.shape[0], 1))
                lr = LinearRegression(fit_intercept=False)
                lr.fit(X_design, X_m[:, target], sample_weight=sample_weight)
                c[target] = lr.coef_[0]
            else:
                coef = np.zeros(reg.coef_.shape)
                lr = LinearRegression(fit_intercept=True)
                lr.fit(X_m[:, pred[pruned_idx]], X_m[:, target], sample_weight=sample_weight)
                coef[pruned_idx] = lr.coef_
                B[target, predictors] = coef
                c[target] = lr.intercept_

        self._adjacency_matrix = B
        self._intercept = c

        return self

    def _logistic_prediction(self, data, params):
        return 1 / (1 + np.exp(-np.dot(sm.add_constant(data), params)))

    def _extract_partial_orders(self, pk):
        """Extract partial orders from prior knowledge."""
        path_pairs = np.array(np.where(pk == 1)).transpose()
        no_path_pairs = np.array(np.where(pk == 0)).transpose()

        # Check for inconsistencies in pairs with path
        check_pairs = np.concatenate([path_pairs, path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            if len(pairs[counts > 1]) > 0:
                raise ValueError(
                    f"The prior knowledge contains inconsistencies at the following indices: {pairs[counts > 1].tolist()}"
                )

        # Check for inconsistencies in pairs without path.
        # If there are duplicate pairs without path, they cancel out and are not ordered.
        check_pairs = np.concatenate([no_path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            check_pairs = np.concatenate([no_path_pairs, pairs[counts > 1]])
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            no_path_pairs = pairs[counts < 2]

        check_pairs = np.concatenate([path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) == 0:
            # If no pairs are extracted from the specified prior knowledge,
            return check_pairs

        pairs = np.unique(check_pairs, axis=0)
        return pairs[:, [1, 0]]  # [to, from] -> [from, to]

    def _residual(self, xi, xj):
        """The residual when xi is regressed on xj."""
        return xi - (np.cov(xi, xj, bias=True)[0, 1] / np.var(xj)) * xj

    def _entropy(self, u):
        """Calculate entropy using the maximum entropy approximations."""
        k1 = 79.047
        k2 = 7.4129
        gamma = 0.37457
        return (1 + np.log(2 * np.pi)) / 2 - k1 * (
            np.mean(np.log(np.cosh(u))) - gamma) ** 2 - k2 * (np.mean(u * np.exp((-(u ** 2)) / 2))) ** 2

    def _diff_mutual_info(self, xi_std, xj_std, ri_j, rj_i):
        """Calculate the difference of the mutual informations."""
        return (self._entropy(xj_std) + self._entropy(ri_j / np.std(ri_j))) - (
            self._entropy(xi_std) + self._entropy(rj_i / np.std(rj_i))
        )

    def _search_candidate_top_down(self, U):
        """Search for candidate features"""
        # If no prior knowledge is specified, nothing to do.
        if self._Aknw is None:
            return U, []

        # Apply prior knowledge in a strong way
        if not self._apply_prior_knowledge_softly:
            if len(self._partial_orders) != 0:
                Uc = [i for i in U if i not in self._partial_orders[:, 1]]
                return Uc, []
            else:
                return U, []

        # Find exogenous features
        Uc = []
        for j in U:
            index = U[U != j]
            if self._Aknw[j][index].sum() == 0:
                Uc.append(j)

        # Find endogenous features, and then find candidate features
        if len(Uc) == 0:
            U_end = []
            for j in U:
                index = U[U != j]
                if np.nansum(self._Aknw[j][index]) > 0:
                    U_end.append(j)

            # Find sink features
            for i in U:
                index = U[U != i]
                if self._Aknw[index, i].sum() == 0:
                    U_end.append(i)
            Uc = [i for i in U if i not in set(U_end)]

        # make V^(j)
        Vj = []
        for i in U:
            if i in Uc:
                continue
            if self._Aknw[i][Uc].sum() == 0:
                Vj.append(i)
        return Uc, Vj

    def _search_causal_order_top_down(self, X, U, min_samples=0):
        """Search the causal ordering from top to bottom."""
        Uc, Vj = self._search_candidate_top_down(U)
        if len(Uc) == 1:
            return Uc[0]

        M_list = []
        for i in Uc:
            M = 0
            for j in U:
                if i != j:
                    X_m = X[:, [i, j]].copy()
                    X_m = X_m[~np.isnan(X_m).any(axis=1)]
                    if X_m.shape[0] < min_samples:
                        return -1
                    xi_std = (X_m[:, 0] - np.mean(X_m[:, 0])) / np.std(X_m[:, 0])
                    xj_std = (X_m[:, 1] - np.mean(X_m[:, 1])) / np.std(X_m[:, 1])
                    ri_j = (
                        xi_std
                        if i in Vj and j in Uc
                        else self._residual(xi_std, xj_std)
                    )
                    rj_i = (
                        xj_std
                        if j in Vj and i in Uc
                        else self._residual(xj_std, xi_std)
                    )
                    M += np.min([0, self._diff_mutual_info(xi_std, xj_std, ri_j, rj_i)]) ** 2
            M_list.append(-1.0 * M)
        return Uc[np.argmax(M_list)]



import os

import numpy as np
import pandas as pd
from lingam.abic_lingam import ABICLiNGAM
from lingam.utils._mggd import MGGD


def test_fit_success():
    d = 4
    B_true = np.zeros((d, d), dtype=float)
    B_true[0, 1] = 1.0  # x0 -> x1
    B_true[1, 2] = -1.5  # x1 -> x2
    B_true[2, 3] = 1.0  # x2 -> x3

    omega_true = np.array(
        [
            [1.2, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.6],  # x1 <-> x3
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.6, 0.0, 1.0],
        ],
        dtype=float,
    )

    n = 100
    beta_shape = 3.0

    mggd = MGGD(np.zeros(d), omega_true, beta_shape)
    eps = mggd.rvs(size=n)

    A = np.eye(d) - B_true
    X = eps @ np.linalg.inv(A)
    X -= X.mean(axis=0, keepdims=True)

    model = ABICLiNGAM(beta=beta_shape)
    model.fit(X)

    am = model.adjacency_matrix_
    cm = model.coefficient_matrix_
    ecm = model.error_covariance_matrix_


def test_fit_invalid_data():

    try:
        model = ABICLiNGAM(beta=-1.0)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(lam=-0.5)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(acyc_order=0)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(acyc_order=1.5)
    except TypeError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(min_causal_effect=-0.1)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(min_error_covariance=-0.1)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(max_outer=0)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(tol_h=0.0)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(rho_max=0.0)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(inner_start=0)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(inner_growth=-1)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model = ABICLiNGAM(inner_tol=0.0)
    except ValueError as e:
        pass
    else:
        raise AssertionError

    model = ABICLiNGAM()

    try:
        model.fit(np.array([1, 2, 3]))
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model.fit(np.array([[1], [2], [3]]))
    except ValueError as e:
        pass
    else:
        raise AssertionError

    try:
        model.fit(pd.DataFrame())
    except ValueError as e:
        pass
    else:
        raise AssertionError


def test_bootstrap_success():
    d = 4
    B_true = np.zeros((d, d), dtype=float)
    B_true[0, 1] = 1.0  # x0 -> x1
    B_true[1, 2] = -1.5  # x1 -> x2
    B_true[2, 3] = 1.0  # x2 -> x3

    omega_true = np.array(
        [
            [1.2, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.6],  # x1 <-> x3
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.6, 0.0, 1.0],
        ],
        dtype=float,
    )

    n = 100
    beta_shape = 3.0
    mggd = MGGD(np.zeros(d), omega_true, beta_shape)
    eps = mggd.rvs(size=n)

    A = np.eye(d) - B_true
    X = eps @ np.linalg.inv(A)
    X -= X.mean(axis=0, keepdims=True)

    model = ABICLiNGAM(beta=beta_shape)
    result = model.bootstrap(X, n_sampling=2)

    am_samples = result.adjacency_matrices_
    cm_samples = result.coefficient_matrices_
    ecm_samples = result.error_covariance_matrices_



import numpy as np
import pandas as pd
import pytest
from lingam import MultiGroupRESIT
from sklearn.ensemble import RandomForestRegressor


def create_data(n=100, seed=0):
    np.random.seed(seed)

    # causal direction: x0 --> x1 --> x3
    X_list = []
    for _ in range(3):
        x0 = np.random.uniform(size=n)
        x1 = 2.0 * x0 + np.random.uniform(size=n)
        x2 = np.random.uniform(size=n)
        x3 = 4.0 * x1 + np.random.uniform(size=n)
        X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])
        X_list.append(X)

    return X_list


def test_fit_success():
    X_list = create_data()

    reg = RandomForestRegressor(max_depth=4, random_state=0)
    model = MultiGroupRESIT(regressor=reg)
    model.fit(X_list)

    assert len(model.adjacency_matrices_) == 3
    assert len(model.causal_order_) == 4

    model.estimate_total_effect(X_list[0], 0, 0)
    model.get_error_independence_p_values(X_list[0])


def test_fit_invalid():
    X_list = create_data()

    reg = RandomForestRegressor(max_depth=4, random_state=0)

    # Not array data
    with pytest.raises(ValueError):
        model = MultiGroupRESIT(regressor=reg)
        model.fit(1)

    # Invalid number of groups
    with pytest.raises(ValueError):
        model = MultiGroupRESIT(regressor=reg)
        model.fit(X_list[:1])

    # Invalid number of groups
    with pytest.raises(ValueError):
        model = MultiGroupRESIT(regressor=reg)
        model.fit(
            [
                X_list[0][["x0", "x1", "x2"]],
                X_list[1][["x0", "x1"]],
                X_list[2][["x0", "x1", "x2"]],
            ]
        )

    # Invalid regressor
    with pytest.raises(ValueError):
        model = MultiGroupRESIT(regressor=None)
        model.fit(X_list)

    # Invalid regressor
    dummy_reg = lambda x: x
    with pytest.raises(ValueError):
        model = MultiGroupRESIT(regressor=dummy_reg)
        model.fit(X_list)

    # Invalid alpha
    with pytest.raises(ValueError):
        model = MultiGroupRESIT(regressor=reg, alpha=-1)
        model.fit(X_list)


def test_bootstrap_success():
    X_list = create_data()

    reg = RandomForestRegressor(max_depth=4, random_state=0)
    model = MultiGroupRESIT(regressor=reg)
    result_list = model.bootstrap(X_list, n_sampling=5)
    result = result_list[0]
    assert len(result.adjacency_matrices_) == 5
    assert len(result.total_effects_) == 5

    # No argument
    cdc = result.get_causal_direction_counts()

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)

    # get_probabilities
    probs = result.get_probabilities()

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

    # get_paths
    paths = result.get_paths(1, 2)


def test_bootstrap_invalid_data():
    X_list = create_data()
    reg = RandomForestRegressor(max_depth=4, random_state=0)
    model = MultiGroupRESIT(regressor=reg)

    # Invalid argument: bootstrap(n_sampling=-1)
    with pytest.raises(ValueError):
        result_list = model.bootstrap(X_list, n_sampling=-1)

    # Invalid argument: bootstrap(n_sampling='3')
    with pytest.raises(ValueError):
        result_list = model.bootstrap(X_list, n_sampling="3")

    result_list = model.bootstrap(X_list, n_sampling=5)

    # Invalid argument: get_causal_direction_counts(n_directions=-1)
    with pytest.raises(ValueError):
        result_list[0].get_causal_direction_counts(n_directions=-1)

    # Invalid argument: get_causal_direction_counts(n_directions='3')
    with pytest.raises(ValueError):
        result_list[0].get_causal_direction_counts(n_directions="3")

    # Invalid argument: get_causal_direction_counts(min_causal_effect=-1.0)
    with pytest.raises(ValueError):
        result_list[0].get_causal_direction_counts(min_causal_effect=-1.0)

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags=-1)
    with pytest.raises(ValueError):
        result_list[0].get_directed_acyclic_graph_counts(n_dags=-1)

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags='3')
    with pytest.raises(ValueError):
        result_list[0].get_directed_acyclic_graph_counts(n_dags="3")

    # Invalid argument: get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    with pytest.raises(ValueError):
        result_list[0].get_directed_acyclic_graph_counts(min_causal_effect=-1.0)

    # Invalid argument: get_probabilities(min_causal_effect=-1.0)
    with pytest.raises(ValueError):
        result_list[0].get_probabilities(min_causal_effect=-1.0)

    # Invalid argument: get_total_causal_effects(min_causal_effect=-1.0)
    with pytest.raises(ValueError):
        result_list[0].get_total_causal_effects(min_causal_effect=-1.0)

    # Invalid argument: get_paths(min_causal_effect=-1.0)
    with pytest.raises(ValueError):
        result_list[0].get_paths(0, 1, min_causal_effect=-1.0)


def test_prior_knowledge_success():
    X_list = create_data()

    reg = RandomForestRegressor(max_depth=4, random_state=0)

    # prior knowledge: nothing
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )
    model = MultiGroupRESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X_list)
    co = model.causal_order_

    # prior knowledge: x1 is exogenous
    pk = np.array(
        [
            [0, -1, -1, -1],
            [0, 0, 0, 0],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )
    model = MultiGroupRESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X_list)

    # prior knowledge: x0 is sink
    pk = np.array(
        [
            [0, -1, -1, -1],
            [0, 0, -1, -1],
            [0, -1, 0, -1],
            [0, -1, -1, 0],
        ]
    )
    model = MultiGroupRESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X_list)

    # prior knowledge: x2-->x3 has path
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, 0],
            [-1, -1, 1, 0],
        ]
    )
    model = MultiGroupRESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X_list)

    # prior knowledge: x1-->x3 does not have path
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, 0, -1, 0],
        ]
    )
    model = MultiGroupRESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X_list)
    co = model.causal_order_


def test_prior_knowledge_invalid():
    # causal direction: x0 --> x1 --> x3
    X_list = create_data(1000)

    reg = RandomForestRegressor(max_depth=4, random_state=0)

    # prior knowledge: invalid
    pk = np.array(
        [
            [0, -1, 1],
            [-1, 0, -1],
            [-1, -1, 0],
            [-1, -1, -1],
        ]
    )
    with pytest.raises(ValueError):
        model = MultiGroupRESIT(regressor=reg, prior_knowledge=pk)
        model.fit(X_list)

    # prior knowledge: inconsistent
    pk = np.array(
        [
            [0, 1, -1, -1],
            [1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )
    with pytest.raises(ValueError):
        model = MultiGroupRESIT(regressor=reg, prior_knowledge=pk)
        model.fit(X_list)



import numpy as np
from scipy.stats import gamma
from scipy.special import digamma
from scipy.optimize import root_scalar


class MGGD:
    """Multivariate Generalized Gaussian Distribution (MGGD)"""

    def __init__(self, mean, cov, beta=1, tolerance=1e-6):
        """Initializes the MGGD with given parameters.

        Parameters
        ----------
        mean : array-like, shape (n_features,)
            Mean vector of the distribution.
        cov : array-like, shape (n_features, n_features)
            Covariance matrix of the distribution.
        beta : float, optional (default=1)
            Shape parameter of the distribution.
        tolerance : float, optional (default=1e-6)
            Tolerance for convergence in parameter estimation.
        """
        self.mean = np.array(mean)
        self.cov = np.array(cov)
        self.beta = beta
        self.tolerance = tolerance

        n_features = self.mean.shape[0]

        if self.cov.shape[0] != n_features or self.cov.shape[1] != n_features:
            raise ValueError("cov must be a square matrix matching the length of mean.")
        if not np.allclose(self.cov, self.cov.T):
            raise ValueError("cov must be a symmetric matrix.")
        if np.any(np.linalg.eigvals(self.cov) < self.tolerance * np.max(np.abs(np.linalg.eigvals(self.cov)))):
            raise ValueError("cov must be a symmetric positive definite matrix.")
        if self.beta <= 0:
            raise ValueError("beta must be a positive value.")

    def rvs(self, size):
        """Generates random samples from the MGGD.

        Parameters
        ----------
        size : int
            Number of samples to generate.

        Returns
        -------
        samples : array-like, shape (size, n_features)
            Generated samples.
        """
        p = self.mean.shape[0]
        eigvals, eigvecs = np.linalg.eigh(self.cov)
        A = eigvecs @ np.diag(np.sqrt(eigvals)) @ eigvecs.T

        r = gamma.rvs(a=p / (2 * self.beta), scale=2, size=size) ** (
            1 / (2 * self.beta)
        )
        U = np.random.multivariate_normal(np.zeros(p), np.eye(p), size)
        U_norm = np.linalg.norm(U, axis=1, keepdims=True)
        U = U / U_norm

        samples = (r[:, np.newaxis] * (U @ A.T)) + self.mean
        return samples


class MGGDEstimator:
    """Estimator for Multivariate Generalized Gaussian Distribution (MGGD) parameters"""

    def __init__(self):
        pass

    def fit(self, X, eps=1e-6):
        """Estimates the parameters of the MGGD from data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Input data.
        eps : float, optional (default=1e-6)
            Small value to avoid division by zero.

        Returns
        -------
        mggd : MGGD
            Estimated MGGD object.
        """
        if X.ndim == 1:
            X = X.reshape(-1, 1)

        n, p = X.shape
        mean = np.mean(X, axis=0)
        X_centered = X - mean

        cov = np.eye(p)
        beta = 0.1
        beta_1 = np.inf

        k = 0
        while np.abs(beta - beta_1) > eps and k < 1000:
            k += 1
            cov_k = np.zeros((p, p))
            inv_cov = np.dot(X_centered, np.linalg.inv(cov))
            u = np.sum(inv_cov * X_centered, axis=1)

            for i in range(n):
                cov_k += u[i] ** (beta - 1) * np.outer(X_centered[i], X_centered[i])

            cov = cov_k / n
            cov = (p / np.sum(np.diag(cov))) * cov

            beta_1 = beta
            beta = self._estimate_beta(u, beta, p, eps)

        m = (beta / (p * n) * np.sum(u**beta)) ** (1 / beta)
        cov *= m

        return MGGD(mean=mean, cov=cov, beta=beta, tolerance=eps)

    def _estimate_beta(self, u, beta, p, eps):
        """Estimates the shape parameter beta using root finding."""
        N = len(u)

        def equation(z):
            term1 = p * N / (2 * np.sum(u**z)) * np.sum(np.log(u + eps) * u**z)
            term2 = p * N / (2 * z) * (np.log(2) + digamma(p / (2 * z)))
            term3 = N
            term4 = p * N / (2 * z) * np.log(z / (p * N) * np.sum(u**z + eps))
            return term1 - term2 - term3 - term4

        bracket = [eps, 2 * np.ceil(beta)]
        f_a, f_b = equation(bracket[0]), equation(bracket[1])

        while f_a * f_b > 0:
            if f_a > 0:
                bracket[0] /= 2
            else:
                bracket[1] *= 2
            f_a, f_b = equation(bracket[0]), equation(bracket[1])

        result = root_scalar(equation, bracket=bracket, method="brentq")
        return result.root
